# -*- coding: utf-8 -*-
import os
import sys

from django.core.management.base import BaseCommand
from django.core.management.base import CommandError
from django.contrib.admin.options import IncorrectLookupParameters
from django.core.exceptions import FieldError
from django.contrib.auth.models import User

from ... import settings
from ...engine import process
from ...sessions import REGISTRY
from ...utils import item_by_attr


class Command(BaseCommand):
    help = 'Run minke-sessions.'

    def add_arguments(self, parser):
        parser.add_argument(
            'session',
            nargs='?',
            help='Session to work with.')
        parser.add_argument(
            'model',
            nargs='?',
            help='Model to work with. (Only neccessary if a session '
                 'could be used on multiple models)')
        parser.add_argument(
            '-q', '--url-query',
            help='Filter objects by url-query.')
        parser.add_argument(
            '-f', '--form-data',
            help='Key-value-pairs used for the session-form.')
        parser.add_argument(
            '-o', '--offset',
            type=int,
            help='Offset')
        parser.add_argument(
            '-l', '--limit',
            type=int,
            help='Limit')
        parser.add_argument(
            '-u', '--user',
            help='User to work with.')
        parser.add_argument(
            '-s', '--list-sessions',
            action='store_true',
            help='List sessions. But do nothing.')
        parser.add_argument(
            '-i',
            '--list-items',
            action='store_true',
            help='List all items to run a session with. But do nothing.')

    def print_usage_and_quit(self, error=None):
        if error:
            print('\033[1;31m' + '[ERROR] ' + str(error) + '\033[0m')
            print()
        self.print_help('manage.py', 'minkesessions')
        sys.exit(1 if error else 0)

    def get_user(self, options, session_cls):
        user = os.environ.get('MINKE_CLI_USER', settings.MINKE_CLI_USER)
        user = options['user'] or user
        try:
            user = User.objects.get(username=user)
        except User.DoesNotExist:
            msg = 'User does not exist: {}'.format(user)
            raise CommandError(msg)
        
        if not user.has_perms(session_cls.permissions):
            msg = 'User is not allowed to run {}: {}'
            raise CommandError(msg.format(session_cls.__name__, user))

        return user

    def get_session_cls(self, options):
        session = options['session']
        if not session:
            msg = 'Missing session-argument'
            raise CommandError(msg)

        try:
            session_cls = REGISTRY[session]
        except KeyError:
            msg = 'Unknown session: {}'.format(session)
            raise CommandError(msg)

        return session_cls

    def get_model_cls(self, options, session_cls):
        model = options['model']

        if model:
            model_cls = item_by_attr(session_cls.work_on, '__name__', model)
            if not model_cls:
                msg = 'Invalid model for {}: {}'.format(session_cls.__name__, model)
                raise CommandError(msg)
        elif len(session_cls.work_on) == 1:
            model_cls = session_cls.work_on[0]
        else:
            msg = 'You need to specify a model to run {} with.'
            msg = msg.format(session_cls, session_cls.work_on)
            raise CommandError(msg)

        return model_cls

    def get_queryset(self, options, model_cls, user):
        if options['url_query']:
            queryset = self.get_changelist_queryset(options, model_cls, user)
        else:
            queryset = model_cls.objects.all()

        # slicing the queryset
        offset = options['offset']
        limit = options['limit']
        if type(offset) is type(limit) is int:
            limit = offset + limit

        try:
            queryset = queryset[offset:limit]
        except AssertionError:
            msg = 'Invalid slicing: [{}:{}]'.format(offset, limit)
            raise CommandError(msg)

        # FIXME: OFFSET- and LIMIT-statements does not work in subqueries,
        # which we use to get the host-query. Therefor we need a workaround:
        if type(offset) is int or type(limit) is int:
            ids = list(queryset.values_list('id', flat=True))
            queryset = model_cls.objects.filter(id__in=ids)

        return queryset

    def get_changelist_queryset(self, options, model_cls, user):
        from django.contrib import admin
        from django.test import RequestFactory
        from django.urls import reverse

        url_query = options['url_query']

        # get a request-instance with session-middleware
        model_label = model_cls._meta.label_lower.replace('.', '_')
        url_pattern = 'admin:' + model_label + '_changelist'
        url = reverse(url_pattern) + '?' + url_query
        request_factory = RequestFactory()
        request = request_factory.get(url)
        request.user = user

        # get a changelist and return its queryset
        modeladmin = admin.site._registry[model_cls]
        try:
            changelist = modeladmin.get_changelist_instance(request)
        except (IncorrectLookupParameters, FieldError):
            msg = 'Invalid url-query: {}'.format(url_query)
            raise CommandError(msg)
        return changelist.get_queryset(request)

    def get_form_data(self, options, session_cls):
        form_cls = session_cls.form
        if not session_cls.form: return dict()

        # form-data passed via command-line?
        form_data = options['form_data']
        if form_data:
            try:
                form_data = eval('dict({})'.format(form_data))
                form = form_cls(form_data)
                assert form.is_valid()
            except AssertionError:
                msg = 'Invalid form-data: {}\n'.format(form_data)
                for field, error in form.errors.items():
                    msg += '{}: {}'.format(field, error[0])
                raise CommandError(msg)
            except Exception as error:
                msg = 'Invalid form-data: {}\n{}'.format(form_data, error)
                raise CommandError(msg)

        # otherwise prompt for it
        else:
            form = form_cls()
            form_data = dict()
            fields = form.visible_fields()
            for field in fields:
                if field.help_text: print(field.help_text)
                form_data[field.name] = input(field.name + ': ')
            form = form_cls(form_data)
            while not form.is_valid():
                for field, error in form.errors.items():
                    if error: print(error[0])
                    form_data[field] = input(field + ': ')
                form = form_cls(form_data)

        # got valid form-data now
        return form.cleaned_data

    def handle(self, *args, **options):
        REGISTRY.reload()

        if options['list_sessions']:
            for session_cls in REGISTRY.values():
                print(session_cls.__name__)
            return

        try:
            session_cls = self.get_session_cls(options)
        except CommandError as err:
            self.print_usage_and_quit(err)

        try:
            user = self.get_user(options, session_cls)
        except CommandError as err:
            self.print_usage_and_quit(err)

        try:
            model_cls = self.get_model_cls(options, session_cls)
        except CommandError as err:
            self.print_usage_and_quit(err)

        try:
            queryset = self.get_queryset(options, model_cls, user)
        except CommandError as err:
            self.print_usage_and_quit(err)

        try:
            form_data = self.get_form_data(options, session_cls)
        except CommandError as err:
            self.print_usage_and_quit(err)

        if options['list_items']:
            for obj in queryset: print(obj)
        else:
            process(session_cls, queryset, form_data, user, console=True)



# -*- coding: utf-8 -*-
# Generated by Django 1.11.12 on 2019-03-26 16:48
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import minke.messages
import minke.utils


class Migration(migrations.Migration):

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('minke', '0002_auto_20180619_1703'),
    ]

    operations = [
        migrations.CreateModel(
            name='BaseMessage',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('level', models.CharField(choices=[('info', 'info'), ('warning', 'warning'), ('error', 'error')], max_length=128)),
                ('text', models.TextField()),
                ('html', models.TextField()),
            ],
        ),
        migrations.CreateModel(
            name='HostGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=255, unique=True)),
                ('comment', models.TextField(blank=True, null=True)),
            ],
            options={
                'ordering': ['name'],
            },
        ),
        migrations.CreateModel(
            name='MinkeSession',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('session_name', models.CharField(max_length=128)),
                ('session_verbose_name', models.CharField(max_length=128)),
                ('session_description', models.TextField(blank=True, null=True)),
                ('session_status', models.CharField(choices=[('success', 'success'), ('warning', 'warning'), ('error', 'error')], max_length=128)),
                ('session_data', minke.utils.JSONField(blank=True, null=True)),
                ('minkeobj_id', models.PositiveIntegerField()),
                ('current', models.BooleanField(default=True)),
                ('proc_status', models.CharField(choices=[('initialized', 'initialized'), ('running', 'running'), ('done', 'done'), ('aborted', 'aborted')], max_length=128)),
                ('start_time', models.DateTimeField(blank=True, null=True)),
                ('end_time', models.DateTimeField(blank=True, null=True)),
                ('run_time', models.DurationField(blank=True, null=True)),
                ('minkeobj_type', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AlterModelOptions(
            name='host',
            options={'ordering': ['name']},
        ),
        migrations.RenameField(
            model_name='host',
            old_name='host',
            new_name='name',
        ),
        migrations.RemoveField(
            model_name='host',
            name='hoststring',
        ),
        migrations.RemoveField(
            model_name='host',
            name='locked',
        ),
        migrations.RemoveField(
            model_name='host',
            name='port',
        ),
        migrations.RemoveField(
            model_name='host',
            name='user',
        ),
        migrations.AddField(
            model_name='host',
            name='comment',
            field=models.TextField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='host',
            name='lock',
            field=models.CharField(blank=True, max_length=20, null=True),
        ),
        migrations.AddField(
            model_name='host',
            name='username',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='host',
            name='verbose_name',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AlterField(
            model_name='host',
            name='hostname',
            field=models.CharField(blank=True, max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='basemessage',
            name='session',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='messages', to='minke.MinkeSession'),
        ),
        migrations.CreateModel(
            name='Message',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=(minke.messages.ProxyMixin, 'minke.basemessage'),
        ),
        migrations.AddField(
            model_name='host',
            name='group',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to='minke.HostGroup'),
        ),
        migrations.CreateModel(
            name='PreMessage',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('minke.message',),
        ),
        migrations.CreateModel(
            name='ExceptionMessage',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('minke.premessage',),
        ),
        migrations.CreateModel(
            name='ExecutionMessage',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('minke.premessage',),
        ),
        migrations.CreateModel(
            name='TableMessage',
            fields=[
            ],
            options={
                'proxy': True,
                'indexes': [],
            },
            bases=('minke.premessage',),
        ),
    ]



# -*- coding: utf-8 -*-

from rest_framework.exceptions import APIException


class InvalidMinkeSetup(Exception):
    """
    Baseclass for all exceptions concerning the minke-setup
    """


class SessionRegistrationError(InvalidMinkeSetup):
    """
    Exception for failing session-registration.
    """
    MSG = '`{}` couldn\'t be registerd: '

    def __init__(self, session, msg):
        msg = self.MSG.format(session.__name__) + msg
        super().__init__(msg)


class InvalidURLQuery(APIException):
    status_code = 400
    default_detail = 'Invalid url-query.'
    default_code = 'invalid_urlquery'



# -*- coding: utf-8 -*-

import logging
import functools
from collections import OrderedDict
from fabric2.runners import Result

from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.utils.text import camel_case_to_spaces
from django.dispatch import Signal

from .models import Host
from .models import MinkeModel
from .models import MinkeSession
from .models import BaseMessage
from .forms import CommandForm
from .messages import PreMessage
from .messages import TableMessage
from .messages import ExecutionMessage
from .exceptions import InvalidMinkeSetup
from .exceptions import SessionRegistrationError
from .utils import FormatDict


logger = logging.getLogger(__name__)


class RegistryDict(OrderedDict):
    """
    A reload-able session-registry.
    """
    reload_sessions = Signal(providing_args=['session_name'])

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._static_sessions = None

    def reload(self, session_name=None):
        """
        Reload the registry by running the factories.
        """
        # We backup the static-sessions and reset the registry before each
        # reload. This way the reload algorithms doesn't have to unregister
        # obsolete sessions.
        if self._static_sessions:
            self.clear()
            self.update(self._static_sessions)
        else:
            self._static_sessions = self.copy()

        # no reloading needed for static sessions
        if session_name and session_name in self:
            return

        # trigger the reload signal
        self.reload_sessions.send(sender=self.__class__, session_name=session_name)


REGISTRY = RegistryDict()


class SessionGroup:
    """
    A group of session - displayed as optgroups in the select-widget.
    Create a group and use it as decorator for Sessions::

        my_group = SessionGroup('My Group')

        @my_group
        class MySession(Session):
            pass
    """
    def __init__(self, name=None):
        self.name = name

    def __call__(self, cls, name=None):
        cls.group = name or self.name
        return cls


class SessionRegistration(type):
    """
    metaclass for Sessions that implements session-registration
    """
    def __new__(cls, name, bases, dct):
        # Setting the abstract-attr explicitly avoids its inheritance.
        dct['abstract'] = dct.get('abstract', False)
        return super().__new__(cls, name, bases, dct)

    def __init__(cls, classname, bases, attrs):
        super().__init__(classname, bases, attrs)
        if not cls.abstract:
            cls.register()
            if cls.auto_permission:
                cls.add_permission()

    def register(cls):
        """
        Register the session-class.
        """
        # some sanity-checks
        if cls.__name__ in REGISTRY:
            msg = 'A session with that name was already registered.'
            raise SessionRegistrationError(cls, msg)

        if not cls.work_on:
            msg = 'At least one minke-model must be specified.'
            raise SessionRegistrationError(cls, msg)

        for model in cls.work_on:
            try:
                assert(model == Host or issubclass(model, MinkeModel))
            except (TypeError, AssertionError):
                msg = '{} is no minke-model.'.format(model)
                raise SessionRegistrationError(cls, msg)

        if issubclass(cls, SingleCommandSession) and not cls.command:
            msg = 'SingleCommandSession needs to specify an command.'
            raise SessionRegistrationError(cls, msg)

        if issubclass(cls, CommandChainSession) and not cls.commands:
            msg = 'CommandChainSession needs to specify commands.'
            raise SessionRegistrationError(cls, msg)

        if issubclass(cls, SessionChain) and not cls.sessions:
            msg = 'SessionChain needs to specify sessions.'
            raise SessionRegistrationError(cls, msg)

        # set verbose-name if missing
        if not cls.verbose_name:
            cls.verbose_name = camel_case_to_spaces(cls.__name__)

        # register session
        REGISTRY[cls.__name__] = cls

    def _get_permission(cls):
        codename = 'run_{}'.format(cls.__name__.lower())
        name = 'Can run {}'.format(cls.__name__)
        lookup = 'minke.{}'.format(codename)
        return codename, name, lookup

    def create_permission(cls):
        """
        Create a run-permission for this session-class.
        """
        content_type = ContentType.objects.get_for_model(MinkeSession)
        codename, name, lookup = cls._get_permission()
        permission, created = Permission.objects.update_or_create(
            codename=codename,
            content_type=content_type,
            defaults=dict(name=name))
        return permission, created

    def delete_permission(cls):
        codename, name, lookup = cls._get_permission()
        try:
            Permission.objects.get(codename=codename).delete()
        except Permission.DoesNotExist:
            pass
        else:
            cls.permissions = tuple(set(cls.permissions) - set((lookup,)))

    def add_permission(cls):
        codename, name, lookup = cls._get_permission()
        cls.permissions += (lookup,)


def protect(method):
    """
    Decorator for session-methods to defer their interrupt.
    """
    @functools.wraps(method)
    def wrapper(obj, *args, **kwargs):
        # are we already protected?
        if obj._busy:
            return method(obj, *args, **kwargs)
        # otherwise protect the method-call by setting the busy-flag
        obj._busy = True
        result = method(obj, *args, **kwargs)
        # if interruption was deferred now is the time to raise it
        if obj._stopped:
            raise KeyboardInterrupt
        obj._busy = False
        return result

    return wrapper


class Session(metaclass=SessionRegistration):
    """Base-class for all session-classes.

    All session-classes must inherit from Session. By defining a subclass of
    Session the subclass will be implicitly registered as session-class and also
    a run-permission will be created for it. To prevent this behavior use an
    abstract session by setting :attr:`.abstract` to True.

    Each session will be instantiated with a fabric-:doc:`fabric:api/connection`
    and an object of :class:`~.models.MinkeSession`. The connection-object
    provides the remote-access, while the minkesession is the database-
    representation of a specific session running for a specific
    :class:`minkemodel-object <.models.MinkeModel>`.

    For a session-class to be useful you at least has to define the
    :meth:`.process`-method and add one or more :class:`~.models.MinkeModel` to
    :attr:`.work_on`-attribute.

    """

    abstract = True
    """
    An abstract session-class won't be registered itself. This is useful if your
    session-class should be a base-class for other sessions.

    Abstract session-classes can be registered manually by calling its
    classmethod :meth:`~.SessionRegistration.register`::

        MySession.register()

    This won't add a run-permission-lookup-string to :attr:`.permissions`. To do so
    use the classmethod :meth:`.SessionRegistration.add_permission`::

        MySession.add_permission()
    """

    verbose_name = None
    """Display-name for sessions."""

    group = None
    """
    Group-name used as optgroup-tag in the select-widget.
    Best practice to group sessions is to use a :class:`.SessionGroup`.
    """

    work_on = tuple()
    """Tuple of minke-models. Models the session can be used with."""

    permissions = tuple()
    """
    Tuple of permission-strings. To be able to run a session a user must have
    all the permissions listed. The strings should have the following format:
    "<app-label>.<permission's-codename>.
    """

    auto_permission = True
    """
    If True a lookup-string for a session-specific run-permission will be
    automatically added to :attr:`.permissions`.

    Note
    ----
    To avoid database-access on module-level we won't create the permission itself.
    Once you setup your sessions you could create run-permissions for all sessions
    using the api-command::

        $ ./manage.py minkeadm --create-permissions
    """

    form = None
    """
    An optional form that will be rendered before the session will be
    processed. The form-data will be accessible within the session as the
    data-property. Use it if the session's processing depends on additional
    user-input-data.

    Instead of setting the form-attribute you can also directly overwrite
    :meth:`.get_form`.
    """

    confirm = False
    """
    If confirm is true, the admin-site asks for a user-confirmation before
    processing a session, which also allows to review the objects the session
    was revoked with.
    """

    invoke_config = dict()
    """
    Session-specific fabric- and invoke-configuration-parameters which will
    be used to initialize a :class:`fabric-connection <fabric.connection.Connection>`.
    The keys must be formatted in a way that is accepted by
    :meth:`~.fabrictools.FabricConfig.load_snakeconfig`.

    See also the documentation for the configuration of
    :doc:`fabric <fabric:concepts/configuration>` and
    :doc:`invoke <invoke:concepts/configuration>`.
    """

    parrallel_per_host = False
    """
    Allow parrallel processing of multiple celery-tasks on a single host.
    If multiple minke-objects are associated with the same host all tasks
    running on them would be processed in a serial manner by default. This is
    to protect the ressources of the host-system. If you want to allow parrallel
    processing of multiple celery-tasks on a single host set parrallel_per_host
    to True.

    Note
    ----
    To perform parrallel task-execution on a single host we make use of celery's
    chords-primitive, which needs a functioning result-backend to be configured.
    Please see the :ref:`celery-documentation <chord-important-notes>`
    for more details.
    """

    def __init__(self, con, db, minkeobj=None):
        """Session's init-method.

        Parameters
        ----------
        con : obj of :class:`fabric.connection.Connection`
        db : obj of :class:`~.models.MinkeSession`
        minkeobj : obj of :class:`~.models.Minkeobj` (optional)
            Only required if you want to initialize a session out of another
            session and let it work on a different minkeobj.
        """
        self._c = con
        self._db = db
        self._minkeobj = minkeobj
        self._stopped = False
        self._busy = False
        self.start = db.start
        self.end = db.end

    @classmethod
    def get_form(cls):
        """
        Return :attr:`.form` by default.

        Overwrite this method if you need to setup your form-class dynamically.
        """
        return cls.form

    @property
    def c(self):
        """
        Refers to the :class:`fabric.connection.Connection`-object the session
        was initialized with.
        """
        return self._c

    @property
    def minkeobj(self):
        """
        Refers to :attr:`.models.MinkeSession.minkeobj`.
        """
        return self._minkeobj or self._db.minkeobj

    @property
    def status(self):
        """
        Refers to :attr:`.models.MinkeSession.session_status`.
        """
        return self._db.session_status

    @property
    def data(self):
        """
        Refers to :attr:`.models.MinkeSession.session_data`.
        This model-field holds all the data that comes from :attr:`.form`.
        """
        return self._db.session_data

    def stop(self, *arg, **kwargs):
        """Interrupt the session's processing.

        This method could be called twice. The first time it will initiate a
        soft interruption which means a current remote-process won't be
        interrupted. The session will be stopped subsequently.
        If it is called meanwhile a second time, the session will be killed
        immediately.

        Note
        ----
        It seems that there is no chance to interrupt a shell-process
        started by fabric if no pty is in use.
        fabric.runners.Remote.send_interrupt says::

            ... in v1, we just reraised the KeyboardInterrupt unless a PTY was
            present; this seems to have been because without a PTY, the
            below escape sequence is ignored, so all we can do is immediately
            terminate on our end...

        Thus killing a session makes most sense if it has the run.pty-config
        in use. Otherwise you just will be disconnected from the remote-process.
        """
        if not self._busy or self._stopped:
            raise KeyboardInterrupt
        else:
            self._stopped = True

    def process(self):
        """
        Real work is done here...

        This is the place for your own session-code.
        """
        raise NotImplementedError('Your session must define a process-method!')

    def add_msg(self, msg, level=None):
        """
        Add a message.

        Parameters
        ----------
        msg
            You could either pass an instance of any :mod:`message-class<.messages>`,
            or any type the different message-classes are initiated with:

            * a string for a :class:`~messages.PreMessage`

            * a tuple or list for a :class:`~messages.TableMessage`

            * an object of :class:`~fabric.runners.Result` for a
              :class:`~messages.ExecutionMessage`

        level : string or bool (optional)
            This could be one of 'info', 'warning' or 'error'. If you pass a
            bool True will be 'info' and False will be 'error'.
        """
        if isinstance(msg, str):
            msg = PreMessage(msg, level)
        elif isinstance(msg, tuple) or isinstance(msg, list):
            msg = TableMessage(msg, level)
        elif isinstance(msg, Result):
            msg = ExecutionMessage(msg, level)
        elif isinstance(msg, BaseMessage):
            pass
        self._db.messages.add(msg, bulk=False)

    def set_status(self, status, update=True):
        """
        Set session-status. Pass a valid session-status or a bool.

        Parameters
        ----------
        status : string or bool
            Valid :attr:`status <models.MinkeSession.SESSION_STATES>` as string
            or True for 'success' and False for 'error'.
        update : bool (optional)
            If True the session-status could only be raised. Lower values as
            current will be ignored.
        """
        states = dict(self._db.SESSION_STATES)
        if type(status) == bool:
            status = 'success' if status else 'error'
        elif status.lower() in states.keys():
            status = status.lower()
        else:
            msg = 'session-status must be one of {}'.format(states.keys())
            raise InvalidMinkeSetup(msg)

        if not self.status or not update or states[self.status] < states[status]:
            self._db.session_status = status

    def format_cmd(self, cmd):
        """
        Use the :attr:`.data` and the fields of the :attr:`.minkeobj` as
        parameters for :func:`format` to prepare the given command.

        Parameters
        ----------
        cmd : string
            a format-string

        Returns
        -------
        string
            The formatted command.
        """
        cmd = cmd.format_map(FormatDict(self.data))
        cmd = cmd.format_map(FormatDict(vars(self.minkeobj)))
        return cmd

    @protect
    def run(self, cmd, **invoke_params):
        """
        Run a command.

        Basically call :meth:`~fabric.connection.Connection.run` on the
        :class:`~fabric.connection.Connection`-object with the given command
        and invoke-parameters.

        Additionally save the :class:`~invoke.runners.Result`-object as an
        :class:`.models.CommandResult`-object.

        Parameters
        ----------
        cmd : string
            The shell-command to be run.
        **invoke_params (optional)
            Parameters that will be passed to
            :meth:`~fabric.connection.Connection.run`

        Returns
        -------
        object of :class:`.models.CommandResult`
        """
        result = self.c.run(cmd, **invoke_params)
        self._db.commands.add(result, bulk=False)
        return result

    @protect
    def frun(self, cmd, **invoke_params):
        """
        Same as :meth:`.run`, but use :meth:`~.format_cmd` to prepare the
        command-string.
        """
        return self.run(self.format_cmd(cmd), **invoke_params)

    @protect
    def xrun(self, cmd, **invoke_params):
        """
        Same as :meth:`.frun`, but also add a
        :class:`~.messages.ExecutionMessage` and update the session-status.
        """
        result = self.frun(cmd, **invoke_params)
        self.add_msg(result)
        self.set_status(result.status)
        return result

    @protect
    def update_field(self, field, cmd, regex=None, **invoke_params):
        """
        Running a command and update a field of :attr:`~.minkeobj`.

        Assign either result.stdout or if available the first matched
        regex-group. If result.failed is True or result.stdout is empty
        or the given regex does not match, the field is updated with None.
        In this case an error-message will be added.

        Parameters
        ----------
        field : string
            Name of the field that should be updated.
        cmd : string
            The shell-command to be run.
        regex: string (optional)
            A regex-pattern the :class:`.CommandResult` will be initialized with.
        **invoke_params (optional)
            Parameters that will be passed to
            :meth:`~fabric.connection.Connection.run`

        Returns
        -------
        bool
            False if the field was updated with None. True otherwise.

        Raises
        ------
        AttributeError
            If the given field does not exists on :attr:`.minkeobj`.
        """
        # is field a minkeobj-attribute?
        try:
            getattr(self.minkeobj, field)
        except AttributeError as e:
            raise e

        result = self.frun(cmd, **invoke_params)

        if regex and result.validate(regex):
            try:
                value = result.match.group(1)
            except IndexError:
                value = result.stdout
        elif not regex and result.ok and result.stdout:
            value = result.stdout
        else:
            self.add_msg(result, 'error')
            self.set_status('warning')
            value = None

        setattr(self.minkeobj, field, value)
        return bool(value)


class SingleCommandSession(Session):
    """
    An abstract :class:`.Session`-class to execute a single command.

    If you want your session to execute a single command simply create a
    subclass of SingleCommandSession and overwrite the
    :attr:`.command`-attribute. The command will be executed with
    :meth:`~.Session.xrun`.

    Example
    -------
    ::

        class MyDrupalModel(models.Model):
            root = models.CharField(max_length=255)

        class MySession(SingleCommandSession):
            work_on = (MyDrupalModel,)
            command = 'drush --root={root} cache-clear all'
    """
    abstract = True

    command = None
    """Shell-command to be executed."""

    def process(self):
        self.xrun(self.command)


class CommandFormSession(SingleCommandSession):
    """
    Same as class:`.SingleCommandSession` but rendering a TextField to enter
    the command.

    Example
    -------
    ::

        class MySession(CommandFormSession):
            work_on = (MyModel,)
    """
    abstract = True
    form = CommandForm
    command = '{cmd}'


class CommandChainSession(Session):
    """
    An abstract :class:`.Session`-class to execute a sequence of commands.

    The commands will be invoked one after another. If one of the commands
    return with a state defined in :attr:`.break_states` no further commands
    will be executed.

    Example
    -------
    ::

        class MySession(CommandChainSession):
            work_on = (MyServer,)
            commands = (
                'a2ensite mysite'
                'apachectl configtest',
                'service apache2 reload')
    """
    abstract = True

    commands = tuple()
    """tuple of shell-commands"""

    break_states = ('error',)
    """
    tuple of :attr:`.models.CommandResult.status` on which the session will
    be interrupted
    """

    def process(self):
        for cmd in self.commands:
            result = self.xrun(cmd)
            if result.status in self.break_states:
                break


class SessionChain(Session):
    """
    An abstract :class:`.Session`-class to execute a sequence of sessions.

    If you have some sessions you want to be able to process separately or as
    a sequence you could make use of a :class:`.SessionChain`.

    All sessions have to work with the same :attr:`~.Session.minkeobj`. If one
    of the sessions ends up with a status defined in :attr:`.break_states` no
    further sessions will be executed.

    Example
    -------
    ::

        class MySession(SessionChain):
            work_on = (MyServer,)
            sessions = (
                UpdateSQLServer,
                RestartSQLServer,
                UpdateApache,
                RestartApache)
    """
    abstract = True

    sessions = tuple()
    """tuple of :class:`.Session`"""

    break_states = ('error',)
    """
    tuple of :attr:`.Session.status` on which further processing will be
    skipped.
    """

    def process(self):
        for cls in self.sessions:
            session = cls(self.c, self._db, self.minkeobj)
            msg = 'Run {}'.format(session.verbose_name)
            self.add_msg(msg)
            session.process()
            if session.status in self.break_states:
                msg = '{} finished with status {}'.format(session.verbose_name, session.status)
                self.add_msg(msg, 'error')
                break



# -*- coding: utf-8 -*-

import sys, traceback, textwrap

from django.utils.html import escape
from .models import BaseMessage
from .settings import MINKE_MESSAGE_WRAP


# We declare the Meta-class whithin a mixin.
# Otherwise the proxy-attribute won't be inherited by child-classes.
class ProxyMixin(object):
    class Meta:
        proxy = True


class Message(ProxyMixin, BaseMessage):
    def __init__(self, data, level=None):
        super().__init__()
        self.text = self.get_text(data)
        self.html = self.get_html(data)
        self.level = self.get_level(data, level)

    def get_level(self, data, level):
        """
        Normalize message-level.
        """
        levels = dict(self.LEVELS).keys()
        if level is None:
            # info as default-level
            return 'info'
        if isinstance(level, bool):
            # True as info and False as error
            return 'info' if level else 'error'
        elif isinstance(level, str) and level.lower() in levels:
            # level as it was given
            return level.lower()
        else:
            raise ValueError('invalid message-level: {}'.format(level))

    def get_text(self, data):
        """
        Derive message-text from data.
        """
        return data

    def get_html(self, data):
        """
        Derive message-html from data.
        """
        return escape(data)


class PreMessage(Message):
    def get_html(self, data):
        data = self.get_text(data)
        wrapped = list()
        for line in data.splitlines():
            wrapped += textwrap.wrap(line, MINKE_MESSAGE_WRAP)
        return '<pre>{}</pre>'.format(escape('\n'.join(wrapped)))


class TableMessage(PreMessage):
    def get_text(self, data):
        widths = dict()
        rows = list()
        spacer = '    '
        for row in data:
            for i, col in enumerate(row):
                if widths.get(i, -1) < len(col):
                    widths[i] = len(col)
        for row in data:
            ljust_row = [s.ljust(widths[i]) for i, s in enumerate(row)]
            rows.append(spacer.join(ljust_row))
        return '\n'.join(rows)


class ExecutionMessage(PreMessage):
    def get_level(self, data, level):
        # if a level-argument was passed just use it
        if not level is None:
            return super().get_level(data, level)
        # otherwise derive the level from result-characteristics
        elif data.failed:
            return 'error'
        elif data.stderr:
            return 'warning'
        else:
            return 'info'

    def _get_chunk(self, prefix, lines):
        wrapped = list()
        prefixed = list()
        for line in lines:
            wrapped += textwrap.wrap(line, MINKE_MESSAGE_WRAP - 10)
        for line in wrapped:
            prefixed.append(prefix.ljust(10) + line)
            prefix = str()
        return prefixed

    def get_text(self, data):
        lines = list()
        prefix = 'code[{}]'.format(data.return_code)
        lines += self._get_chunk(prefix, data.command.splitlines())
        lines += self._get_chunk('stdout:', data.stdout.splitlines())
        lines += self._get_chunk('stderr:', data.stderr.splitlines())
        return '\n'.join(lines)


class ExceptionMessage(PreMessage):
    def __init__(self, level='error', print_tb=False):
        type, value, tb = sys.exc_info()

        if print_tb:
            data = traceback.format_exception(type, value, tb)
        else:
            data = traceback.format_exception_only(type, value)

        data = str().join(data)
        super().__init__(data, level)



# -*- coding: utf-8 -*-

from rest_framework import serializers

from .models import MinkeSession
from .models import BaseMessage


class MessageSerializer(serializers.ModelSerializer):
    class Meta:
        model = BaseMessage
        fields = ('level', 'html')


class SessionSerializer(serializers.ModelSerializer):
    messages = MessageSerializer(many=True, read_only=True)

    class Meta:
        model = MinkeSession
        fields = ('id', 'minkeobj_id', 'session_status', \
                  'proc_status', 'proc_info', 'messages', 'is_done')
        read_only_fields = fields



# Generated by Django 2.2.4 on 2019-08-10 14:12

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('minke', '0004_auto_20190703_1319'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='minkesession',
            name='task_id',
        ),
        migrations.AddField(
            model_name='minkesession',
            name='pid',
            field=models.IntegerField(blank=True, help_text='Process-ID of the celery-task that run the session.', null=True, verbose_name='PID'),
        ),
    ]



# -*- coding: utf-8 -*-

from django import forms
from django.forms.widgets import Select
from django.utils.translation import gettext_lazy as _
from .utils import prepare_shell_command


class MinkeForm(forms.Form):
    session = forms.CharField(required=True, widget=forms.HiddenInput())
    select_across = forms.BooleanField(
        widget=forms.HiddenInput(),
        required=False,
        initial=False)
    minke_form = forms.BooleanField(
        widget=forms.HiddenInput(),
        required=False,
        initial=True)

    class Media:
        css = {'all': ('minke/css/minke_form.css',)}


class PassphraseForm(forms.Form):
    connect_kwargs_passphrase = forms.CharField(
        label=_('Passphrase'),
        help_text=_('Passphrase to decrypt an ssh-private-key.'),
        max_length=100,
        widget=forms.PasswordInput(attrs={
            'placeholder': 'Passphrase',
            'autofocus': '',
            'required': ''})
        )


class CommandField(forms.CharField):
    widget = forms.Textarea
    def clean(self, value):
        value = super().clean(value)
        return prepare_shell_command(value)


class CommandForm(forms.Form):
    cmd = CommandField(
        label=_('Command'),
        help_text=_('Command to execute. Use with care!'),
    )


class ExtendedSelect(Select):
    """
    Select that allows to pass option-specific extra-attributes.
    """
    def __init__(self, attrs=None, extra_attrs=None, choices=()):
        super().__init__(attrs, choices)
        self.extra_attrs = extra_attrs or dict()

    def create_option(self, *args, **kwargs):
        option = super().create_option(*args, **kwargs)
        option['attrs'].update(self.extra_attrs.get(option['label'], dict()))
        return option


class SessionSelectForm(forms.Form):
    session = forms.ChoiceField(label=_('Session:'), widget=ExtendedSelect)



# -*- coding: utf-8 -*-

import re
from django.dispatch import receiver

from minke.sessions import CommandFormSession
from minke.sessions import SingleCommandSession
from minke.sessions import CommandChainSession
from minke.sessions import REGISTRY
from .forms import CommandForm


class BaseCommandSession(SingleCommandSession):
    abstract = True
    model = None
    model_id = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        obj = self.model.objects.get(pk=self.model_id)
        self.command = obj.cmd


class BaseCommandChainSession(CommandChainSession):
    abstract = True
    model = None
    model_id = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        obj = self.model.objects.get(pk=self.model_id)
        self.commands = (c.cmd for c in obj.commands.all())


class BaseCommandChoiceSession(CommandFormSession):
    abstract = True
    model = None
    model_id = None
    form = CommandForm

    @classmethod
    def get_form(cls):
        obj = cls.model.objects.get(pk=cls.model_id)
        choices = ((c.id, str(c)) for c in obj.commands.all())
        cls.form.base_fields['cmd'].choices = choices
        return cls.form

    def format_cmd(self, cmd):
        cmd_obj = self.model.commands.get(pk=self.data['cmd'])
        return super().format_cmd(cmd_obj.cmd)


@receiver(REGISTRY.reload_sessions)
def session_factory(sender, **kwargs):

    # FIXME: Prevent those inline-imports!
    from .models import Command
    from .models import CommandGroup

    # do we have a session-name and does it match our regex?
    session_name = kwargs.get('session_name', None)
    pattern = r'^(Command|CommandGroup)_(\d+)'
    match = re.match(pattern, session_name or '')

    # if we have no session-name, register all command-sessions
    if not session_name:
        for obj in Command.objects.filter(active=True):
            session = obj.as_session()
            session.register()
            session.add_permission()
            session.create_permission()

        for obj in CommandGroup.objects.filter(active=True):
            session = obj.as_session()
            session.register()
            session.add_permission()
            session.create_permission()

    # if we have a session-name that matches, register this session only
    elif match:
        clsname, id = match.groups()
        cls = Command if clsname == 'Command' else CommandGroup
        cmd = cls.objects.get(pk=id)
        session = cmd.as_session()
        session.register()
        session.add_permission()
        session.create_permission()



# -*- coding: utf-8 -*-

from celery import chain
from celery import group

from .messages import Message
from .messages import ExceptionMessage
from .models import MinkeSession
from .tasks import process_session
from .tasks import cleanup


def process(session_cls, queryset, session_data, user,
            fabric_config=None, wait=False, console=False):
    """
    Initiate and run celery-tasks.
    """
    MinkeSession.objects.clear_currents(user, queryset)
    hosts = queryset.get_hosts()
    lock = hosts.filter(disabled=False).get_lock()

    # group sessions by hosts
    session_groups = dict()
    for minkeobj in queryset.select_related_hosts():
        host = minkeobj.get_host()

        session = MinkeSession()
        session.init(user, minkeobj, session_cls, session_data)

        # Skip disabled or locked hosts...
        if host.disabled:
            msg = '{}: Host is disabled.'.format(minkeobj)
            session.messages.add(Message(msg, 'error'), bulk=False)
            session.cancel()
            if console: session.prnt()
        elif host.lock and host.lock != lock:
            msg = '{}: Host is locked.'.format(minkeobj)
            session.messages.add(Message(msg, 'error'), bulk=False)
            session.cancel()
            if console: session.prnt()

        # otherwise group sessions by hosts...
        else:
            if host not in session_groups:
                session_groups[host] = list()
            session_groups[host].append(session)

    # Stop here if no valid hosts are left...
    if not session_groups: return

    # merge fabric-config and invoke-config
    config = session_cls.invoke_config.copy()
    config.update(fabric_config or dict())


    # run celery-tasks...
    results = list()
    for host, sessions in session_groups.items():

        # get process_session_signatures for all sessions
        signatures = [process_session.si(host.id, s.id, config) for s in sessions]

        # To support parrallel execution per host we wrap the signatures in a group.
        # NOTE: Since we append the cleanup-task the construct is essentially the
        # same as a chord which is not supported by all result-backends (s. celery-docs).
        if session_cls.parrallel_per_host:
            signatures = [group(*signatures)]

        # append the cleanup-task
        signatures.append(cleanup.si(host.id))

        try:
            result = chain(*signatures).delay()

        # NOTE: celery-4.2.1 fails to raise an exception if rabbitmq is
        # down or no celery-worker is running at all... hope for 4.3.x
        except process_session.OperationalError:
            host.release_lock()
            for session in sessions:
                session.add_msg(ExceptionMessage())
                session.cancel()
                if console: session.prnt(session)

        else:
            results.append((result, (s.id for s in sessions)))


    # print sessions in cli-mode as soon as they are ready...
    if console:
        print_results = results[:]
        while print_results:
            # try to find a ready result...
            try: result, session_ids = next((r for r in print_results if r[0].ready()))
            except StopIteration: continue
            # reload session-objects
            sessions = MinkeSession.objects.filter(id__in=session_ids)
            # print and remove list-item
            for session in sessions: session.prnt()
            print_results.remove((result, session_ids))

    # evt. wait till all tasks finished...
    elif wait:
        for result, sessions in results:
            result.wait()

    # At least call forget on every result - in case a result-backend is in use
    # that eats up ressources to store result-data...
    for result, sessions in results:
        try: result.forget()
        except NotImplementedError: pass



# -*- coding: utf-8 -*-

import logging
import signal

from socket import error as SocketError
from socket import gaierror as GaiError
from paramiko.ssh_exception import SSHException
from fabric2 import Connection
from invoke.exceptions import Failure
from invoke.exceptions import ThreadException
from invoke.exceptions import UnexpectedExit
from celery import shared_task

from . import settings
from .models import Host
from .models import MinkeSession
from .sessions import REGISTRY
from .messages import ExceptionMessage
from .settings import MINKE_HOST_CONFIG
from .settings import MINKE_FABRIC_CONFIG


logger = logging.getLogger(__name__)


class SessionProcessor:
    """
    Process sessions.
    """
    def __init__(self, host_id, session_id, session_config):
        host = Host.objects.select_related('group').get(pk=host_id)
        hostname = host.hostname or host.name
        config = MINKE_FABRIC_CONFIG.clone()
        self.host = host

        # At first try to load the hostgroup- and host-config...
        for obj in (host.group, host):
            if not obj or not obj.config: continue
            if obj.config in MINKE_HOST_CONFIG:
                config.load_snakeconfig(MINKE_HOST_CONFIG[obj.config])
            else:
                msg = 'Invalid MINKE_HOST_CONFIG for {}'.format(obj)
                logger.warning(msg)

        # At least load the session-config...
        config.load_snakeconfig(session_config or dict())

        # Initialize the connection...
        self.con = Connection(hostname, host.username, host.port, config=config)

        # Initialize the session...
        self.minke_session = MinkeSession.objects.get(pk=session_id)
        REGISTRY.reload(self.minke_session.session_name)
        session_cls = REGISTRY[self.minke_session.session_name]
        self.session = session_cls(self.con, self.minke_session)

    def run(self):
        try:
            started = self.session.start()
            if not started: return

            try:
                self.session.process()

            # Since task-interruption could happen all along between
            # session.start() and session.end() we handle it in the outer
            # try-construct.
            except KeyboardInterrupt:
                raise

            # paramiko- and socket-related exceptions (ssh-layer)
            except (SSHException, GaiError, SocketError):
                self.session.end(failure=True)
                self.session.add_msg(ExceptionMessage())

            # invoke-related exceptions (shell-layer)
            except (Failure, ThreadException, UnexpectedExit):
                self.session.end(failure=True)
                self.session.add_msg(ExceptionMessage())

            # other exceptions raised by process (which is user-code)
            except Exception:
                self.session.end(failure=True)
                exc_msg = ExceptionMessage(print_tb=True)
                logger.error(exc_msg.text)
                if settings.MINKE_DEBUG:
                    self.session.add_msg(exc_msg)
                else:
                    self.session.add_msg('An error occurred.', 'error')

            else:
                self.session.end()

        # task-interruption
        except KeyboardInterrupt:
            self.session.end()

        # at least close the ssh-connection
        finally:
            self.con.close()


@shared_task(bind=True)
def process_session(task, host_id, session_id, config):
    """
    Task for session-processing.
    """
    processor = SessionProcessor(host_id, session_id, config)
    signal.signal(signal.SIGUSR1, processor.session.stop)
    processor.run()

@shared_task
def cleanup(host_id):
    """
    Task to release the host's lock.
    """
    Host.objects.get(pk=host_id).release_lock()



# -*- coding: utf-8 -*-

from django.contrib import admin
from .models import Command
from .models import CommandGroup
from .models import CommandOrder


class CommandOrderInline(admin.TabularInline):
    model = CommandOrder
    fields = ('command',)
    extra = 1


class BaseCommandAdmin(admin.ModelAdmin):
    list_editable = ('active',)
    ordering = ('label',)
    list_filter = ('minketypes',)

    def minketypes_view(self, obj):
        types = (ct.model_class().__name__ for ct in obj.minketypes.all())
        return ', '.join(types)
    minketypes_view.short_description = 'Minke-Models'


@admin.register(Command)
class CommandAdmin(BaseCommandAdmin):
    list_display = ('cmd', 'label', 'minketypes_view', 'active')
    search_fields = ('label', 'cmd')


@admin.register(CommandGroup)
class CommandGroupAdmin(BaseCommandAdmin):
    inlines = (CommandOrderInline,)
    list_display = ('label', 'minketypes_view', 'active')
    search_fields = ('label',)



# -*- coding: utf-8 -*-

import functools

from fabric2 import Connection
from fabric2.config import Config
from fabric2.runners import Remote

from .exceptions import InvalidMinkeSetup
from .models import CommandResult


class FabricConfig(Config):
    """
    A subclass of fabric's Config-class.
    Add a load_snakeconfig-method, that takes a plain dict and parses its
    snake-case-keys to fit into the nested default-config-structure.
    """
    def __init__(self, *args, **kwargs):
        # NOTE:
        # fabric has a rather dodgy way to check for optional config-files:
        #
        # > # Typically means 'no such file', so just note & skip past.
        # > except IOError as e:
        # >     # TODO: is there a better / x-platform way to detect this?
        # >     if "No such file" in e.strerror:
        # >         err = "Didn't see any {}, skipping."
        # >         debug(err.format(filepath))
        # >     else:
        # >         raise
        #
        # To prevent an accidentally raised IOError we catch it and re-try to
        # initialize Config with `lazy=True` (which means `Config` won't look for
        # config-files at all).
        # FIXME: Initialization fails when building the docs cause sphinx seems
        # to somehow localize the error-message. Find a way to not let sphinx do
        # that, and remove the __init__-hacking.
        try:
            super().__init__(*args, **kwargs)
        except IOError:
            kwargs['lazy'] = True
            super().__init__(*args, **kwargs)

    def load_snakeconfig(self, configdict):
        """
        Load a plane configdict as nested config - using the key's
        snake-structure as representation of the nest-logic.

        Two level of recursion are supported. That means something like
        'connect_kwargs_my_special_key' will be applied as
        'config.connect_kwargs.my_special_key'. To obviate ambiguity the key on
        the first level must already exist. Otherwise we raise InvalidMinkeSetup.

        Parameter
        ---------
        configdict : dict
        """
        for param, value in configdict.items():
            snippets = param.split('_')
            key1 = key2 = None

            for i, key in enumerate(snippets):
                if '_'.join(snippets[:i+1]) in self:
                    key1 = '_'.join(snippets[:i+1])
                    key2 = '_'.join(snippets[i+1:])
                    break

            if not key1:
                msg = 'Invalid fabric-config-parameter: {}'.format(param)
                raise InvalidMinkeSetup(msg)

            # apply config-data
            if not key2:
                self[key1] = value
            else:
                if not self[key1]:
                    # FIXME: normal dict does not support the attr-api.
                    self[key1] = dict()
                self[key1][key2] = value


class FabricRemote(Remote):
    """
    A subclass of fabric's remote-runner to customize the result-class.
    """
    def generate_result(self, **kwargs):
        kwargs["connection"] = self.context
        return CommandResult(**kwargs)


def protect(method):
    """
    Decorator for session-methods to defer their interrupt.
    """
    @functools.wraps(method)
    def wrapper(obj, *args, **kwargs):
        # are we already protected?
        if obj._busy:
            return method(obj, *args, **kwargs)
        # otherwise protect the method-call by setting the busy-flag
        obj._busy = True
        result = method(obj, *args, **kwargs)
        # if interruption was deferred now is the time to raise it
        if obj._stopped:
            raise KeyboardInterrupt
        obj._busy = False
        return result

    return wrapper


class FabricConnection(Connection):
    """
    Extend fabric's Connection by the minkeobj and the db-session and some
    helper-methods.
    """
    def __init__(self, *args, **kwargs):
        self._db = kwargs.pop('db')
        self._minkeobj = kwargs.pop('minkeobj', None)
        self._stopped = False
        self._busy = False
        super().__init__(self, *args, **kwargs)

    @property
    def minkeobj(self):
        """
        Refers to :attr:`.models.MinkeSession.minkeobj`.
        """
        return self._minkeobj or self._db.minkeobj

    @property
    def status(self):
        """
        Refers to :attr:`.models.MinkeSession.session_status`.
        """
        return self._db.session_status

    def add_msg(self, msg, level=None):
        """
        Add a message.

        Parameters
        ----------
        msg
            You could either pass an instance of any :mod:`message-class<.messages>`,
            or any type the different message-classes are initiated with:
            * a string for a :class:`~messages.PreMessage`
            * a tuple or list for a :class:`~messages.TableMessage`
            * an object of :class:`~fabric.runners.Result` for a
              :class:`~messages.ExecutionMessage`
        level : string or bool (optional)
            This could be one of 'info', 'warning' or 'error'. If you pass a
            bool True will be 'info' and False will be 'error'.
        """
        if isinstance(msg, str):
            msg = PreMessage(msg, level)
        elif isinstance(msg, tuple) or isinstance(msg, list):
            msg = TableMessage(msg, level)
        elif isinstance(msg, Result):
            msg = ExecutionMessage(msg, level)
        elif isinstance(msg, BaseMessage):
            pass
        self._db.messages.add(msg, bulk=False)

    def set_status(self, status, update=True):
        """
        Set session-status. Pass a valid session-status or a bool.

        Parameters
        ----------
        status : string or bool
            Valid :attr:`status <models.MinkeSession.SESSION_STATES>` as string
            or True for 'success' and False for 'error'.
        update : bool (optional)
            If True the session-status could only be raised. Lower values as
            current will be ignored.
        """
        states = dict(self._db.SESSION_STATES)
        if type(status) == bool:
            status = 'success' if status else 'error'
        elif status.lower() in states.keys():
            status = status.lower()
        else:
            msg = 'session-status must be one of {}'.format(states.keys())
            raise InvalidMinkeSetup(msg)

        if not self.status or not update or states[self.status] < states[status]:
            self._db.session_status = status

    def format_cmd(self, cmd):
        """
        Use the :attr:`.data` and the fields of the :attr:`.minkeobj` as
        parameters for :func:`format` to prepare the given command.

        Parameters
        ----------
        cmd : string
            a format-string

        Returns
        -------
        string
            The formatted command.
        """
        cmd = cmd.format_map(FormatDict(self.data))
        cmd = cmd.format_map(FormatDict(vars(self.minkeobj)))
        return cmd

    @protect
    def run(self, cmd, **invoke_params):
        """
        Run a command.

        Basically call :meth:`~fabric.connection.Connection.run` on the
        :class:`~fabric.connection.Connection`-object with the given command
        and invoke-parameters.

        Additionally save the :class:`~invoke.runners.Result`-object as an
        :class:`.models.CommandResult`-object.

        Parameters
        ----------
        cmd : string
            The shell-command to be run.
        **invoke_params (optional)
            Parameters that will be passed to
            :meth:`~fabric.connection.Connection.run`

        Returns
        -------
        object of :class:`.models.CommandResult`
        """
        result = super().run(cmd, **invoke_params)
        self._db.commands.add(result, bulk=False)
        return result

    @protect
    def frun(self, cmd, **invoke_params):
        """
        Same as :meth:`.run`, but use :meth:`~.format_cmd` to prepare the
        command-string.
        """
        return self.run(self.format_cmd(cmd), **invoke_params)

    @protect
    def xrun(self, cmd, **invoke_params):
        """
        Same as :meth:`.frun`, but also add a
        :class:`~.messages.ExecutionMessage` and update the session-status.
        """
        result = self.frun(cmd, **invoke_params)
        self.add_msg(result)
        self.set_status(result.status)
        return result

    @protect
    def update_field(self, field, cmd, regex=None, **invoke_params):
        """
        Running a command and update a field of :attr:`~.minkeobj`.

        Assign either result.stdout or if available the first matched
        regex-group. If result.failed is True or result.stdout is empty
        or the given regex does not match, the field is updated with None.
        In this case an error-message will be added.

        Parameters
        ----------
        field : string
            Name of the field that should be updated.
        cmd : string
            The shell-command to be run.
        regex: string (optional)
            A regex-pattern the :class:`.CommandResult` will be initialized with.
        **invoke_params (optional)
            Parameters that will be passed to
            :meth:`~fabric.connection.Connection.run`

        Returns
        -------
        bool
            False if the field was updated with None. True otherwise.

        Raises
        ------
        AttributeError
            If the given field does not exists on :attr:`.minkeobj`.
        """
        # is field a minkeobj-attribute?
        try:
            getattr(self.minkeobj, field)
        except AttributeError as e:
            raise e

        result = self.frun(cmd, **invoke_params)

        if regex and result.validate(regex):
            try:
                value = result.match.group(1)
            except IndexError:
                value = result.stdout
        elif not regex and result.ok and result.stdout:
            value = result.stdout
        else:
            self.add_msg(result, 'error')
            self.set_status('warning')
            value = None

        setattr(self.minkeobj, field, value)
        return bool(value)



# -*- coding: utf-8 -*-
from django.template import Library
from django.contrib.admin.templatetags.admin_list import result_list


register = Library()


@register.inclusion_tag('minke/change_list_results.html', takes_context=True)
def minke_result_list(context, cl):
    """
    Support rendering of session-info-rows for changelist-results.
    """
    sessions = cl.sessions if hasattr(cl, 'sessions') else cl.result_list
    cntxt = result_list(cl)
    # django1 has problems with RequestConext together with forloop, cycle and
    # nested inclusion-tags.So we use the context as dictonary.
    cntxt.update(context.flatten())
    cntxt['results'] = zip(cntxt['results'], sessions)
    return cntxt


@register.inclusion_tag('minke/session_select.html', takes_context=True)
def minke_session_select(context):
    """
    Render the session-select-form.
    """
    cl = context['cl']
    request = context['request']
    form = cl.model_admin.get_session_select_form(request)
    context['session_select_form'] = form
    context['session_count'] = cl.session_count
    return context


@register.inclusion_tag('minke/session_switch.html', takes_context=True)
def minke_session_switch(context):
    """
    Render the session-switch-bar.
    """
    return context


@register.inclusion_tag('minke/session.html', takes_context=True)
def minke_session(context):
    """
    Render the session-info-row.
    """
    return context


@register.simple_tag(takes_context=True)
def update_url_query(context, *args):
    """
    Update the current url-query.
    """
    query_dict = context['request'].GET.copy()
    args = list(args)
    while args:
        key = args.pop(0)
        if key.startswith('-'):
            if key[1:] in query_dict: del query_dict[key[1:]]
        else:
            query_dict[key] = args.pop(0)
    return '?' + query_dict.urlencode()



# -*- coding: utf-8 -*-

from django.conf import settings
from .fabrictools import FabricConfig
from .fabrictools import FabricRemote


MINKE_DEBUG = getattr(settings, 'MINKE_DEBUG', False)
MINKE_FABRIC_FORM = getattr(settings, 'MINKE_FABRIC_FORM', None)
MINKE_CLI_USER = getattr(settings, 'MINKE_CLI_USER', 'admin')
MINKE_FABRIC_CONFIG = getattr(settings, 'MINKE_FABRIC_CONFIG', FabricConfig())
MINKE_HOST_CONFIG = getattr(settings, 'MINKE_HOST_CONFIG', dict())
MINKE_MESSAGE_WRAP = getattr(settings, 'MINKE_MESSAGE_WRAP', 120)

# set defaults for fabric-/invoke-config
MINKE_FABRIC_CONFIG.run.hide = True
MINKE_FABRIC_CONFIG.run.warn = True
MINKE_FABRIC_CONFIG.runners.remote = FabricRemote

# All config-vars starting with FABRIC_ will be loaded into our fabric-config...
fabric_data = dict([(k[7:].lower(), getattr(settings, k)) for k in dir(settings) if k.startswith('FABRIC')])
MINKE_FABRIC_CONFIG.load_snakeconfig(fabric_data)



# Generated by Django 2.1.7 on 2019-07-03 13:19

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import fabric2.runners
import minke.utils


class Migration(migrations.Migration):

    dependencies = [
        ('minke', '0003_auto_20190326_1648'),
    ]

    operations = [
        migrations.CreateModel(
            name='CommandResult',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('command', models.TextField(help_text='The command which was executed.', verbose_name='Command')),
                ('exited', models.SmallIntegerField(help_text='Exit-status returned by the command.', verbose_name='Exit-status')),
                ('stdout', models.TextField(blank=True, help_text='Standard-output of the command.', null=True, verbose_name='Stdout')),
                ('stderr', models.TextField(blank=True, help_text='Standard-error of the command. (unless the process was invoked via a pty, in which case stderr and stdout are merged into stdout)', null=True, verbose_name='Stderr')),
                ('shell', models.CharField(help_text='The shell binary used for execution.', max_length=128, verbose_name='Shell')),
                ('encoding', models.CharField(help_text='The string encoding used by the local shell environment.', max_length=128, verbose_name='Encoding')),
                ('pty', models.BooleanField(help_text='A boolean describing whether the command was invoked with a pty or not', verbose_name='Pty')),
                ('created_time', models.DateTimeField(auto_now_add=True, help_text='The datetime this command-result were added.', verbose_name='Time of creation')),
            ],
            options={
                'verbose_name_plural': 'Command-Results',
                'ordering': ('session_id', 'created_time'),
                'verbose_name': 'Command-Result',
            },
            bases=(fabric2.runners.Result, models.Model),
        ),
        migrations.AlterModelOptions(
            name='basemessage',
            options={'ordering': ('session_id', 'created_time'), 'verbose_name': 'Message', 'verbose_name_plural': 'Messages'},
        ),
        migrations.AlterModelOptions(
            name='host',
            options={'ordering': ['name'], 'verbose_name': 'Host', 'verbose_name_plural': 'Hosts'},
        ),
        migrations.AlterModelOptions(
            name='hostgroup',
            options={'ordering': ['name'], 'verbose_name': 'Host-Group', 'verbose_name_plural': 'Host-Groups'},
        ),
        migrations.AlterModelOptions(
            name='minkesession',
            options={'ordering': ('minkeobj_type_id', 'minkeobj_id', '-created_time'), 'verbose_name': 'Session', 'verbose_name_plural': 'Sessions'},
        ),
        migrations.AddField(
            model_name='basemessage',
            name='created_time',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, help_text='The datetime this message were added.', verbose_name='Time of creation'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='host',
            name='config',
            field=models.CharField(blank=True, help_text='Use config as specified in MINKE_HOST_CONFIG.', max_length=255, null=True, verbose_name='Fabric-/Invoke-config'),
        ),
        migrations.AddField(
            model_name='host',
            name='port',
            field=models.IntegerField(blank=True, help_text='Port number to connect on the remote host.', null=True, verbose_name='Port number.'),
        ),
        migrations.AddField(
            model_name='hostgroup',
            name='config',
            field=models.CharField(blank=True, help_text='Use config as specified in MINKE_HOST_CONFIG.', max_length=255, null=True, verbose_name='Fabric-/Invoke-config'),
        ),
        migrations.AddField(
            model_name='minkesession',
            name='created_time',
            field=models.DateTimeField(auto_now_add=True, default=django.utils.timezone.now, help_text='Time the session has been initiated.', verbose_name='Created-time'),
            preserve_default=False,
        ),
        migrations.AddField(
            model_name='minkesession',
            name='task_id',
            field=models.CharField(blank=True, help_text='ID of the celery-task that run the session.', max_length=128, null=True, verbose_name='Task-ID'),
        ),
        migrations.AlterField(
            model_name='basemessage',
            name='html',
            field=models.TextField(help_text='Message as HTML', verbose_name='HTML'),
        ),
        migrations.AlterField(
            model_name='basemessage',
            name='level',
            field=models.CharField(choices=[('info', 'info'), ('warning', 'warning'), ('error', 'error')], help_text='Level with which the message were added.', max_length=128, verbose_name='Message-level'),
        ),
        migrations.AlterField(
            model_name='basemessage',
            name='session',
            field=models.ForeignKey(help_text='Session the message belongs to.', on_delete=django.db.models.deletion.CASCADE, related_name='messages', to='minke.MinkeSession', verbose_name='Session'),
        ),
        migrations.AlterField(
            model_name='basemessage',
            name='text',
            field=models.TextField(help_text='Message-Text', verbose_name='Text'),
        ),
        migrations.AlterField(
            model_name='host',
            name='comment',
            field=models.TextField(blank=True, help_text='Something about the host.', null=True, verbose_name='Comment'),
        ),
        migrations.AlterField(
            model_name='host',
            name='disabled',
            field=models.BooleanField(default=False, help_text="Disabled hosts won't be accessed by minke.", verbose_name='Disabled'),
        ),
        migrations.AlterField(
            model_name='host',
            name='group',
            field=models.ForeignKey(blank=True, help_text='The group this host belongs to.', null=True, on_delete=django.db.models.deletion.SET_NULL, to='minke.HostGroup', verbose_name='Hostgroup'),
        ),
        migrations.AlterField(
            model_name='host',
            name='hostname',
            field=models.CharField(blank=True, help_text='HostName could be either a ssh-config-lookup-pattern or a real hostname to log into.', max_length=255, null=True, verbose_name='HostName'),
        ),
        migrations.AlterField(
            model_name='host',
            name='lock',
            field=models.CharField(blank=True, help_text="Locked hosts won't be accessed by minke.To prevent intersection a host will be locked while sessions are executed on it.", max_length=20, null=True, verbose_name='Lock'),
        ),
        migrations.AlterField(
            model_name='host',
            name='name',
            field=models.SlugField(help_text="Unique name of the host. If Hostname is not specified fabric's Connection-class will be initialized with this Name instead. Specifying a Name only could be a sufficient Host-setup if there is a valid ssh_config for it as lookup-pattern.", max_length=128, unique=True, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='host',
            name='username',
            field=models.CharField(blank=True, help_text='User to login as.', max_length=255, null=True, verbose_name='User'),
        ),
        migrations.AlterField(
            model_name='host',
            name='verbose_name',
            field=models.CharField(blank=True, help_text='Verbose Host-Name.', max_length=255, null=True, verbose_name='Verbose Name'),
        ),
        migrations.AlterField(
            model_name='hostgroup',
            name='comment',
            field=models.TextField(blank=True, help_text='Something about the group.', null=True, verbose_name='Comment'),
        ),
        migrations.AlterField(
            model_name='hostgroup',
            name='name',
            field=models.CharField(help_text='Unique group-name.', max_length=255, unique=True, verbose_name='Group-Name'),
        ),
        migrations.AlterField(
            model_name='minkesession',
            name='end_time',
            field=models.DateTimeField(blank=True, help_text='Time the session finished.', null=True, verbose_name='End-time'),
        ),
        migrations.AlterField(
            model_name='minkesession',
            name='proc_status',
            field=models.CharField(choices=[('initialized', 'initialized'), ('running', 'running'), ('completed', 'completed'), ('stopping', 'stopping'), ('stopped', 'stopped'), ('canceled', 'canceled'), ('failed', 'failed')], help_text='Status of session-processing.', max_length=128, verbose_name='Process-status'),
        ),
        migrations.AlterField(
            model_name='minkesession',
            name='run_time',
            field=models.DurationField(blank=True, help_text="Session's runtime.", null=True, verbose_name='Run-time'),
        ),
        migrations.AlterField(
            model_name='minkesession',
            name='session_data',
            field=minke.utils.JSONField(blank=True, help_text='Data coming from a session-form.', null=True, verbose_name="Session's extra-data"),
        ),
        migrations.AlterField(
            model_name='minkesession',
            name='session_description',
            field=models.TextField(blank=True, help_text='Doc-string of the session-class.', max_length=128, null=True, verbose_name="Session's description"),
        ),
        migrations.AlterField(
            model_name='minkesession',
            name='session_name',
            field=models.CharField(help_text='Class-name of the session-class.', max_length=128, verbose_name='Session-name'),
        ),
        migrations.AlterField(
            model_name='minkesession',
            name='session_status',
            field=models.CharField(choices=[('success', 'success'), ('warning', 'warning'), ('error', 'error')], help_text='Mostly set by the session-code itself.', max_length=128, verbose_name='Session-status'),
        ),
        migrations.AlterField(
            model_name='minkesession',
            name='session_verbose_name',
            field=models.CharField(help_text='Verbose-name-attribute of the session-class.', max_length=128, verbose_name="Session's verbose-name"),
        ),
        migrations.AlterField(
            model_name='minkesession',
            name='start_time',
            field=models.DateTimeField(blank=True, help_text='Time the session has been started.', null=True, verbose_name='Start-time'),
        ),
        migrations.AlterField(
            model_name='minkesession',
            name='user',
            field=models.ForeignKey(help_text='User that run this session.', on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL, verbose_name='User'),
        ),
        migrations.AddField(
            model_name='commandresult',
            name='session',
            field=models.ForeignKey(help_text='Session whereas this command where executed.', on_delete=django.db.models.deletion.CASCADE, related_name='commands', to='minke.MinkeSession', verbose_name='Session'),
        ),
    ]



# Generated by Django 2.1.7 on 2019-05-12 17:25

from django.db import migrations, models
import django.db.models.deletion
import minke.contrib.commands.models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Command',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('cmd', models.TextField()),
                ('label', models.CharField(max_length=256)),
                ('description', models.TextField(blank=True)),
                ('active', models.BooleanField(default=True)),
                ('minketypes', models.ManyToManyField(limit_choices_to=minke.contrib.commands.models.get_minketypes, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='CommandGroup',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('label', models.CharField(max_length=256)),
                ('description', models.TextField(blank=True)),
                ('as_options', models.BooleanField(default=False)),
                ('active', models.BooleanField(default=True)),
            ],
        ),
        migrations.CreateModel(
            name='CommandOrder',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('order', models.SmallIntegerField()),
                ('command', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='commands.Command')),
                ('commands', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='commands.CommandGroup')),
            ],
            options={
                'get_latest_by': 'order',
            },
        ),
        migrations.AddField(
            model_name='commandgroup',
            name='commands',
            field=models.ManyToManyField(through='commands.CommandOrder', to='commands.Command'),
        ),
        migrations.AddField(
            model_name='commandgroup',
            name='minketypes',
            field=models.ManyToManyField(limit_choices_to=minke.contrib.commands.models.get_minketypes, to='contenttypes.ContentType'),
        ),
        migrations.AlterUniqueTogether(
            name='commandorder',
            unique_together={('commands', 'order')},
        ),
    ]



# -*- coding: utf-8 -*-

from django.contrib.auth.models import Permission
from django.core.management.base import BaseCommand
from django.core.management.base import CommandError

from ...models import Host
from ...models import MinkeSession
from ...sessions import REGISTRY


class Command(BaseCommand):
    help = 'Some cleanup-actions.'

    def add_arguments(self, parser):
        parser.add_argument(
            '-L', '--release-locks',
            action='store_true',
            help='Release locks on all hosts.')
        parser.add_argument(
            '-C', '--clear-current-sessions',
            action='store_true',
            help='Set current to False for all sessions.')
        parser.add_argument(
            '-S', '--clear-all-sessions',
            action='store_true',
            help='Delete all sessions.')
        parser.add_argument(
            '-s', '--list-sessions',
            action='store_true',
            help='List available sessions.')
        parser.add_argument(
            '-p', '--create-run-permission',
            help='Create permission for a session.')
        parser.add_argument(
            '-P', '--create-run-permissions',
            action='store_true',
            help='Create permissions for all sessions.')
        parser.add_argument(
            '-D', '--delete-run-permissions',
            action='store_true',
            help='Delete all run-permissions.')

    def handle(self, *args, **options):
        if options['release_locks']:
            print(Host.objects.update(lock=None))

        if options['clear_current_sessions']:
            print(MinkeSession.objects.update(current=False))

        if options['clear_all_sessions']:
            print(MinkeSession.objects.all().delete())

        if options['list_sessions']:
            REGISTRY.reload()
            for session_cls in REGISTRY.values():
                print(session_cls.__name__)

        if options['create_run_permission']:
            REGISTRY.reload()
            session = options['create_permission']
            try:
                session_cls = REGISTRY[session]
            except KeyError:
                msg = 'Unknown session: {}'.format(session)
                raise CommandError(msg)
            permission, created = session_cls.create_permission()
            if created:
                print('Created permission: {}'.format(permission))

        if options['create_run_permissions']:
            REGISTRY.reload()
            for session_cls in REGISTRY.values():
                permission, created = session_cls.create_permission()
                if created:
                    print('Created permission: {}'.format(permission))

        if options['delete_run_permissions']:
            print(Permission.objects.filter(codename__startswith='run_').delete())

